import 'dart:async';
import 'dart:io';
import 'package:dio/src/FormData.dart';
import 'package:dio/src/DioError.dart';
import 'package:dio/src/Interceptor.dart';
import 'package:dio/src/Options.dart';
import 'package:dio/src/Response.dart';
import 'package:dio/src/TransFormer.dart';

typedef OnDownloadProgress(int received, int total);

/// A powerful Http client for Dart, which supports Interceptors,
/// Global configuration, FormData, File downloading etc. and Dio is
/// very easy to use.
class Dio {

  /// Create Dio instance with default [Options].
  /// It's mostly just one Dio instance in your application.
  Dio([Options options]){
    if(options==null){
      options=new Options();
    }
    this.options=options;
  }

  /// The Dio version.
  static const version = "0.0.1";

  /// Request config.
  Options options;

  /// The httpClient, you can config proxy
  HttpClient httpClient = new HttpClient();
  var _interceptor = new Interceptor();

  /// Each Dio instance has a interceptor by which you can intercept requests or responses before they are
  /// handled by `then` or `catchError`. the [interceptor] field
  /// contains a [RequestInterceptor] and a [ResponseInterceptor] instance.
  Interceptor get interceptor => _interceptor;

  /// [TransFormer] allows changes to the request/response data before it is sent/received to/from the server
  /// This is only applicable for request methods 'PUT', 'POST', and 'PATCH'.
  TransFormer transFormer = new DefaultTransformer();

  /**
   * Make http request with options.
   *
   * [path] The url path.
   * [data] The request data
   * [options] The request options.
   */
  Future<Response> request(String path, {data, Options options}) async {
    return _checkIfNeedEnqueue(interceptor.request, () {
      _mergeOptions(options);
      options.data = data ?? options.data;
      options.path = path;
      // If user provide a request interceptor, enter the interceptor.
      InterceptorCallback preSend = interceptor.request.onSend;
      if (preSend != null) {
        var ret = preSend(options);
        // Assure the return value type of request interceptor is `Future`.
        if (ret is! Future) {
          ret = new Future.value(ret);
        }
        return ret.then((_options) {
          // If the Future value type is Options, continue the network request.
          if (_options is Options) {
            return _makeRequest(_options);
          } else {
            // Otherwise, use the Future value as the request result.
            // If the return type is Error, we should throw it
            if (_options is Error) {
              throw _assureHttpError(_options);
            }
            return _assureResponse(_options);
          }
        });
      } else {
        // If user don't provide the request interceptor, make request directly.
        return _makeRequest(options);
      }
    });
  }

  _checkIfNeedEnqueue(interceptor, callback()) {
    if (interceptor.locked) {
      return interceptor.enqueue(callback);
    } else {
      return callback();
    }
  }

  Future<Response> _makeRequest(Options options) async {
    try {
      String url = options.path;
      if (!url.startsWith("http")) {
        url = options.baseUrl + url;
        List<String> s = url.split(":/");
        url = s[0] + ':/' + s[1].replaceAll("//", "/");
      }
      String method = options.method.toUpperCase();
      bool isGet = method == "GET";
      if (isGet && options.data is Map) {
        url += (url.contains("?") ? "&" : "?") +
            TransFormer.urlEncodeMap(options.data);
      }
      Uri uri = Uri.parse(url);
      var request = await httpClient.openUrl(method, uri);
      if (!isGet) {
        await _transformData(options, method, request);
      }
      //todo handle timeout
      var response = await request.close();
      var retData = await transFormer.transformResponse(options, response);
      Response ret = new Response(
          data: retData,
          headers: response.headers,
          request: options,
          statusCode: response.statusCode);

      return _checkIfNeedEnqueue(interceptor.response, () {
        if ((response.statusCode >= 200 && response.statusCode < 300) ||
            response.statusCode == 304) {
          return _onSuccess(ret);
        } else {
          var err = new DioError(response: ret,
              message: 'Http status error [${response.statusCode}]');
          return _onError(err);
        }
      });
    } catch (e) {
      return _checkIfNeedEnqueue(interceptor.response, () {
        return _onError(_assureHttpError(e));
      });
    }
  }

  _transformData(Options options, String method,
      HttpClientRequest request) async {
    var data = options.data;
    if (data != null) {
      if ("POST" == method) {
        // Handle the FormData
        if (data is FormData) {
          request.headers.set(
              HttpHeaders.CONTENT_TYPE,
              'multipart/form-data; ${data.boundary}');
          List<int> content = data.bytes();
          //Must set the content-length
          request.contentLength = content.length;
          options.headers.forEach((k, v) => request.headers.set(k, v));
          request.add(content);
          return;
        }
      }
      options.headers[HttpHeaders.CONTENT_TYPE] =
          options.contentType.toString();
      // Call request transformer.
      data = await transFormer.transformRequest(options);
      // Set the headers, must before `request.write`
      options.headers.forEach((k, v) => request.headers.set(k, v));
      request.write(data);
    }
  }

  // Transform current Future status("success" and "error") if necessary
  Future _transFutureStatusIfNecessary(Future future) {
    return future.then((data) {
      // Strictly be a HttpError instance, but we loos the restrictions
      //if (data is HttpError)
      if (data is Error) {
        return reject(data);
      }
      return resolve(data);
    }, onError: (err) {
      if (err is Response) {
        return resolve(err);
      }
      return reject(err);
    });
  }

  Future<Response> _onSuccess(response) {
    if (interceptor.response.onSuccess != null) {
      response = interceptor.response.onSuccess(response) ?? response;
      if (response is! Future) {
        // Assure response is a Future
        response = new Future.value(response);
      }
      return _transFutureStatusIfNecessary(response);
    }

    if (response is! Future) {
      // Assure response is a Future
      response = resolve(response);
    }
    return response;
  }

  Future<Response> _onError(err) {
    if (interceptor.response.onError != null) {
      err = interceptor.response.onError(err) ?? err;
      if (err is! Future) {
        // Assure err is a Future
        err = new Future.error(err);
      }
      return _transFutureStatusIfNecessary(err);
    }
    if (err is! Future) {
      err = new Future.error(err);
    }
    return err;
  }

  _mergeOptions(Options opt) {
    opt.method ??= options.method ?? "GET";
    opt.method = opt.method.toUpperCase();
    opt.headers.addAll(options.headers);
    opt.baseUrl ??= options.baseUrl ?? "";
    opt.timeout ??= options.timeout ?? 0;
    opt.responseType ??= options.responseType ?? ResponseType.JSON;
    opt.data ??= options.data;
    opt.extra.addAll(options.extra);
    opt.contentType ??= options.contentType ?? ContentType.JSON;
  }

  Options _checkOptions(method, options) {
    if (options == null) {
      options = new Options();
    }
    options.method = method;
    return options;
  }

  DioError _assureHttpError(err) {
    if (err is DioError) {
      return err;
    } else if (err is Error) {
      err = new DioError(response: null,
          message: err.toString(), stackTrace: err.stackTrace);
    } else {
      err = new DioError(message: err.toString());
    }
    return err;
  }

  Response _assureResponse(response) {
    if (response is! Response) {
      response = new Response(data: response);
    }
    return response;
  }


  // Assure the final future state is succeed!
  Future<Response> resolve(response) {
    if (response is! Future) {
      response = new Future.value(response);
    }
    return response.then((data) {
      return _assureResponse(data);
    }, onError: (err) {
      // transform "error" to "success"
      return _assureResponse(err);
    });
  }

  // Assure the final future state is failed!
  Future<Response> reject(err) {
    if (err is! Future) {
      err = new Future.error(err);
    }
    return err.then((v) {
      // transform "success" to "error"
      throw _assureHttpError(v);
    }, onError: (e) {
      throw _assureHttpError(e);
    });
  }

  /**
   * Lock the current Dio instance.
   *
   * Dio will enqueue the incoming request tasks instead
   * send them directly when [interceptor.request] is locked.
   *
   */
  lock() {
    interceptor.request.lock();
  }

  /**
   * Unlock the current Dio instance.
   *
   * Dio instance dequeue the request task。
   */
  unlock() {
    interceptor.request.unlock();
  }

  /**
   * Download the file and save it in local.
   *
   * [urlPath]: The file url.
   *
   * [savePath]: The path to save the downloading file later.
   *
   * [onProgress]: The callback to listen downloading progress.
   * please refer to [OnDownloadProgress].
   *
   */
  Future<Response> download(String urlPath, savePath,
      {OnDownloadProgress onProgress, data, Options options}) async {
    options = _checkOptions("GET", options);
    // We set the `responseType` to [ResponseType.STREAM] to retrieve the
    // response stream.
    options.responseType = ResponseType.STREAM;
    Response<HttpClientResponse> response = await request(
        urlPath, data: data, options: options);
    File file = new File(savePath);
    //Create a new Completer to notify the success/error state.
    Completer completer = new Completer();
    Future future = completer.future;
    int received = 0;
    var list = new List<int>();
    // Listen the downloading progress.
    response.data.listen((data) {
      list.addAll(data);
      received += data.length;
      if (onProgress != null) {
        onProgress(received, response.data.contentLength);
      }
    },
        onDone: () {
          // Write full content to file by once.
          file.writeAsBytesSync(list);
          completer.complete();
        },
        onError: (e) => completer.completeError(_assureHttpError(e)),
        cancelOnError: true);

    return future;
  }

  /// Handy method to make http GET request, which is a alias of  [Dio.request].
  Future<Response> get(path, {data, Options options}) {
    return request(path, data: data, options: _checkOptions("GET", options));
  }
  /// Handy method to make http POST request, which is a alias of  [Dio.request].
  Future<Response> post(String path, {data, Options options}) {
    return request(path, data: data, options: _checkOptions("POST", options));
  }
  /// Handy method to make http PUT request, which is a alias of  [Dio.request].
  Future<Response> put(String path, {data, Options options}) {
    return request(path, data: data, options: _checkOptions("PUT", options));
  }
  /// Handy method to make http HEAD request, which is a alias of  [Dio.request].
  Future<Response> head(String path, {data, Options options}) {
    return request(path, data: data, options: _checkOptions("HEAD", options));
  }
  /// Handy method to make http DELETE request, which is a alias of  [Dio.request].
  Future<Response> delete(String path, {data, Options options}) {
    return request(path, data: data, options: _checkOptions("DELETE", options));
  }
  /// Handy method to make http PATCH request, which is a alias of  [Dio.request].
  Future<Response> patch(String path, {data, Options options}) {
    return request(path, data: data, options: _checkOptions("PATCH", options));
  }

}
